#include <iostream>
#include <iomanip>
#include <cmath>
#include <algorithm>
#include <random>
#include <functional>

#include <btas/btas.h>

void Fix_left_gauge (const btas::Tensor<double>& g, btas::Tensor<double>& X)
{
  for(int i = 0; i < X.extent(0); ++i)
    for(int j = 0; j < X.extent(1); ++j)
      for(int k = 0; k < X.extent(2); ++k)
        X(i,j,k) *= g(i);
}

void Fix_right_gauge (const btas::Tensor<double>& g, btas::Tensor<double>& X)
{
  for(int i = 0; i < X.extent(0); ++i)
    for(int j = 0; j < X.extent(1); ++j)
      for(int k = 0; k < X.extent(2); ++k)
        X(i,j,k) *= g(k);
}

void Fix_left_gauge_inverse (const btas::Tensor<double>& g, btas::Tensor<double>& X)
{
  for(int i = 0; i < X.extent(0); ++i)
    for(int j = 0; j < X.extent(1); ++j)
      for(int k = 0; k < X.extent(2); ++k)
        X(i,j,k) /= g(i);
}

void Fix_right_gauge_inverse (const btas::Tensor<double>& g, btas::Tensor<double>& X)
{
  for(int i = 0; i < X.extent(0); ++i)
    for(int j = 0; j < X.extent(1); ++j)
      for(int k = 0; k < X.extent(2); ++k)
        X(i,j,k) /= g(k);
}

void Truncate (
        btas::Tensor<double>& AB,
        btas::Tensor<double>& g,
        btas::Tensor<double>& A,
        btas::Tensor<double>& B,
        size_t M)
{
  btas::Tensor<double> S(0);
  btas::Tensor<double> U(0,0,0);
  btas::Tensor<double> Vt(0,0,0);
  btas::gesvd('S','S',AB,S,U,Vt);

  size_t m = (S.size() > M) ? M : S.size();

  g.resize(m);
  for(int i = 0; i < m; ++i)
    g(i) = S(i);

  A.resize(U.extent(0),U.extent(1),m);
  for(int i = 0; i < U.extent(0); ++i)
    for(int j = 0; j < U.extent(1); ++j)
      for(int k = 0; k < M; ++k)
        A(i,j,k) = U(i,j,k);

  B.resize(m,Vt.extent(1),Vt.extent(2));
  for(int i = 0; i < M; ++i)
    for(int j = 0; j < Vt.extent(1); ++j)
      for(int k = 0; k < Vt.extent(2); ++k)
        B(i,j,k) = Vt(i,j,k);
}

double propagate (
  const btas::Tensor<double>& expH,
        btas::Tensor<double>& A, btas::Tensor<double>& gA,
        btas::Tensor<double>& B, btas::Tensor<double>& gB,
        size_t M)
{
  // Gauge stuffs
  Fix_right_gauge(gB,B);

  // (Ex.1) Here, contract A and B to make 2-site tensor C

  // (Ex.2) Here, contract C and expH to compute a propagated 2-site tensor W

  // Doing SVD
  Truncate(W,gA,A,B,M);

  // Normalizing the gauge of A
  double Norm2 = btas::dot(gA,gA);
  btas::scal(1.0/sqrt(Norm2),gA);

  // Gauge stuffs
  Fix_left_gauge(gA,B);
  Fix_right_gauge_inverse(gB,B);

  // This is approximate energy expectation
  return -log(WNorm2)/CNorm2/2.0;
}

int main () {

  double J  = 1.0;
  size_t M  =  20;

  size_t Nt = 1000;
  double dt = 0.03;

  btas::Tensor<double> expH(2,2,2,2); expH.fill(0.0);

  double expJz = exp(-0.25*J*dt);
  double coshJ = cosh(0.5*J*dt);
  double sinhJ = sinh(0.5*J*dt);

  // (Ex.3) Followings are non-zero component of exp(-H*dt) where H = J*S(i)*S(i+1)
  // Fill the following
  expH(0,0,0,0) = expJz;
  expH(0,1,0,1) =
  expH(1,0,1,0) =
  expH(0,1,1,0) = 
  expH(1,0,0,1) =
  expH(1,1,1,1) =
  // NOTE: index of expH is in the order [bra(i), bra(i+1), ket(i), ket(i+1)]

  std::mt19937 rgen;
  std::uniform_real_distribution<double> dist(-1.0,1.0);

  btas::Tensor<double> A(M,2,M);
  A.generate(std::bind(dist,rgen));

  btas::Tensor<double> gA(M);
  gA.fill(1.0);

  btas::Tensor<double> B(M,2,M);
  B.generate(std::bind(dist,rgen));

  btas::Tensor<double> gB(M);
  gB.fill(1.0);

  std::cout.precision(8);

  for(int i = 0; i < Nt; ++i) {

    double E = 0.0;

    E += propagate(expH,A,gA,B,gB,M)/dt;

    E += propagate(expH,B,gB,A,gA,M)/dt;

    if(i%10 == 9)
    std::cout << "\tStep[" << std::setw(4) << i << "] E = " << std::fixed << std::setw(12) << E << std::endl;

  }

  return 0;
}
